from django.urls import reverse
from rest_framework import status

from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Group, LevelRight, User


class TestMembership(FixtureAPITestCase):
    '''
    Test groups management
    '''

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.unverified = User.objects.create_user('user@address.com', 'P4$5w0Rd')
        cls.admin = User.objects.create_superuser('admin@address.com', 'P4$5w0Rd')

        cls.admin_group = cls.admin.groups.create(name="Admin group", public=True)
        cls.admin_group.memberships.update(level=LevelRight.Admin.level)

    def test_verified_user_group_create(self):
        """
        Anonymous and non verified users should not be able to create a group
        """
        self.client.force_login(self.unverified)
        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:groups-list'), {'name': 'new group'})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.client.logout()
        with self.assertNumQueries(0):
            response = self.client.post(reverse('api:groups-list'), {'name': 'new group'})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_group_wrong_fields(self):
        checks = (
            ({'name': 100 * 'A'}, {'name': ['Ensure this field has no more than 64 characters.']}),
            ({'name': ''}, {'name': ['This field may not be blank.']})
        )
        self.client.force_login(self.user)
        for payload, error in checks:
            response = self.client.post(reverse('api:groups-list'), payload)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertEqual(response.json(), error)

    def test_create_group(self):
        """
        A user creating a group is granted a membership with the maximum privileges level
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:groups-list'), {'name': 'new group', 'public': True})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        group = Group.objects.get(name='new group')
        self.assertEqual(response.json(), {
            'id': str(group.id),
            'name': 'new group',
            'public': True,
            'members_count': 1
        })
        self.assertCountEqual(
            group.memberships.values_list('user', 'level'),
            [(self.user.id, LevelRight.Admin.level)]
        )

    def test_crate_group_duplicated_name(self):
        """
        Assert multiple groups can have a similar name
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:groups-list'), {'name': self.group.name})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_list_groups(self):
        """
        List groups which are public or for which the user is a member
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(reverse('api:groups-list'))
        self.assertDictEqual(response.json(), {
            'count': 2,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                {
                    'id': str(self.admin_group.id),
                    'members_count': 1,
                    'name': 'Admin group',
                    'public': True
                }, {
                    'id': str(self.group.id),
                    'members_count': 3,
                    'name': 'User group',
                    'public': False
                }
            ]
        })

    def test_list_members_requires_member_user(self):
        """
        Only users that belong to a group have the ability to list the members
        Otherwise we return a 404 as usual
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse('api:group-members', kwargs={'pk': str(self.admin_group.id)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_list_members_level_right(self):
        """
        List members of a group with their right corresponding to privileges level
        Members are ordered by display name
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:group-members', kwargs={'pk': str(self.group.id)}))
        read_member, write_member, admin_member = self.group.memberships.order_by('level')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json(), {
            'count': 3,
            'next': None,
            'number': 1,
            'previous': None,
            'results': [
                {
                    'display_name': 'Test user',
                    'email': admin_member.user.email,
                    'id': str(admin_member.user.id),
                    'level': 'admin'
                }, {
                    'display_name': 'Test user read',
                    'email': read_member.user.email,
                    'id': str(read_member.user.id),
                    'level': 'read'
                }, {
                    'display_name': 'Test user write',
                    'email': write_member.user.email,
                    'id': str(write_member.user.id),
                    'level': 'write'
                }
            ]
        })

    def test_list_members_null_right(self):
        """
        Non defined right level should return None
        """
        self.client.force_login(self.user)
        group = self.user.groups.create(name='Another group', through_defaults={'level': 1})
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:group-members', kwargs={'pk': str(group.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json(), {
            'count': 1,
            'next': None,
            'number': 1,
            'previous': None,
            'results': [
                {
                    'display_name': self.user.display_name,
                    'email': self.user.email,
                    'id': str(self.user.id),
                    'level': None
                }
            ]
        })

    def test_retrieve_group_no_member(self):
        """
        A non memeber of a group may not retrieve its details, even if it is public
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse('api:group-details', kwargs={'pk': str(self.admin_group.id)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_retrieve_group_details(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(reverse('api:group-details', kwargs={'pk': str(self.group.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.group.id),
            'members_count': 3,
            'name': self.group.name,
            'public': self.group.public
        })

    def test_update_group_no_admin(self):
        """
        User must be a group administrator to edit its properties
        """
        non_admin = User.objects.filter(
            memberships__group=self.group,
            memberships__level__lt=LevelRight.Admin.level
        ).first()
        self.client.force_login(non_admin)
        with self.assertNumQueries(2):
            response = self.client.patch(reverse('api:group-details', kwargs={'pk': str(self.group.id)}), {'name': 'A'})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'You do not have permission to perform this action.'
        })

    def test_update_group(self):
        """
        Group name and public attributes may be updated by an admin member
        """
        payload = {
            'name': 'Renamed group',
            'public': True,
            'id': 1337,
            'members_count': 42
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse('api:group-details', kwargs={'pk': str(self.group.id)}),
                payload
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.group.id),
            'members_count': 3,
            'name': 'Renamed group',
            'public': True
        })

    def test_delete_group_no_admin(self):
        non_admin = User.objects.filter(
            memberships__group=self.group,
            memberships__level__lt=LevelRight.Admin.level
        ).first()
        non_admin.verified_email = True
        non_admin.save()
        self.client.force_login(non_admin)
        with self.assertNumQueries(4):
            response = self.client.delete(reverse('api:group-details', kwargs={'pk': str(self.group.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'Only members with an admin privilege may update or delete this group.'
        })

    def test_delete_group(self):
        """
        A group admin is allowed to delete the group
        """
        group = Group.objects.create(name='Another group')
        user = group.users.create(
            email='user42@test.test',
            verified_email=True,
            through_defaults={'level': LevelRight.Admin.level}
        )
        self.client.force_login(user)
        with self.assertNumQueries(6):
            response = self.client.delete(reverse('api:group-details', kwargs={'pk': str(group.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Group.DoesNotExist):
            group.refresh_from_db()
        self.assertTrue(User.objects.filter(id=user.id).exists())

    def test_list_user_memberships_requires_login(self):
        """
        User must be logged in to list its memberships
        """
        with self.assertNumQueries(0):
            response = self.client.delete(reverse('api:group-details', kwargs={'pk': str(self.group.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_list_user_memberships(self):
        """
        List groups for which the user is a member
        """
        a_group = self.user.groups.create(name='Another group', through_defaults={'level': LevelRight.Read.level})
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:user-memberships'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 2,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                {
                    'group':
                        {
                            'id': str(a_group.id),
                            'members_count': 1,
                            'name': 'Another group',
                            'public': False
                        },
                    'id': a_group.memberships.get(user=self.user).id,
                    'level': 'read'
                },
                {
                    'group':
                        {
                            'id': str(self.group.id),
                            'members_count': self.group.users.count(),
                            'name': self.group.name,
                            'public': self.group.public
                        },
                    'id': self.group.memberships.get(user=self.user).id,
                    'level': 'admin'
                }
            ]
        })
